# upload_results.py
# Copyright 2020 Roger Marsh
# Licence: See LICENSE.txt (BSD licence)

"""Submit results submission files to the ECF.
"""

import tkinter
import tkinter.ttk
import tkinter.filedialog
import tkinter.messagebox
import urllib.parse
import subprocess
import html.parser
import re
import os

try:
    if subprocess.run(["curl", "-V"], capture_output=True).returncode:
        curl = False
    else:
        curl = True
except:
    curl = False
try:
    import requests
except ModuleNotFoundError:
    requests = None

_AVOID_SCROLLBAR = "avoidscrollbar"
_START_TEXT = " ".join(
    (
        "Fill in all fields and select options, then submit to ECF. ",
        "Right-click for menu.",
    )
)
_DEFAULT_LIVE_URL = "https://www.ecfrating.org.uk/v2/submit/"
_DEFAULT_TEST_URL = "https://www.ecfrating.org.uk/sandbox/submit/"
_EVENT_DETAILS = "EVENT DETAILS"
_EVENT_CODE = "EVENT CODE"
_SUBMISSION_INDEX = "SUBMISSION INDEX"
_EVENT_NAME = "EVENT NAME"
_RESULTS_OFFICER_ADDRESS = "RESULTS OFFICER ADDRESS"

_whitespace_at_end_re = re.compile(r".*\s+\Z", flags=re.DOTALL)
_whitespace_at_start_re = re.compile(r"\A\s+.*", flags=re.DOTALL)
_feedback_player_list_re = re.compile(
    r"".join(
        (
            r"\s+Submitted\s+Players\s+",
            r"(.*?)",
            r"\s+Submitted\s+Games\s+",
        )
    ),
    flags=re.DOTALL,
)
_submission_player_list_re = re.compile(
    r"".join(
        (
            r"\s*#\s*PlayerList\s*",
            r"(#\s*.*?)",
            r"\s*#\s*",
            r"(?:SectionResults|OtherResults|MatchResults)",
            r"\s*=\s*",
        )
    ),
    flags=re.DOTALL,
)
_feedback_number_re = re.compile(r"\s+\d+\.\s+", flags=re.DOTALL)
_submission_pin_re = re.compile(r"#PIN=\d+", flags=re.DOTALL)


# This is in .core.feedback but do not want the dependency yet.
class Feedback(html.parser.HTMLParser):

    """Parse a feedback file."""

    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        self.feedbackdata = []
        self.feedbackstring = ""
        self.feedbacknumbers = None
        self.feedbackplayers = None
        self.submissionpins = None
        self.submissionplayers = None
        self._ignore_data = 0
        self.submission_file_name = None
        self.responsestring = None

    def insert_whitespace(self):
        fbd = self.feedbackdata
        for i in range(len(fbd) - 1, 0, -1):
            if (
                _whitespace_at_start_re.match(fbd[i]) is None
                and _whitespace_at_end_re.match(fbd[i - 1]) is None
            ):
                fbd.insert(i, " ")
        self.feedbackstring = "".join(fbd)

    def find_player_lists(self):
        fbpl = _feedback_player_list_re.search(self.feedbackstring)
        spl = _submission_player_list_re.search(self.feedbackstring)
        if fbpl:
            self.feedbacknumbers = _feedback_number_re.findall(fbpl.group(1))
            self.feedbackplayers = _feedback_number_re.split(fbpl.group(1))
        if spl:
            self.submissionpins = _submission_pin_re.findall(spl.group(1))
            self.submissionplayers = _submission_pin_re.split(spl.group(1))

    def handle_starttag(self, tag, attrs):
        if tag.strip() in {"script", "style"}:
            self._ignore_data += 1

    def handle_endtag(self, tag):
        if tag.strip() in {"script", "style"}:
            self._ignore_data -= 1

    def handle_startendtag(self, tag, attrs):
        super().handle_startendtag(tag, attrs)

    def handle_data(self, tag):
        if self._ignore_data:
            return
        ts = tag.strip()
        if not ts:
            return
        self.feedbackdata.append(ts)

    def handle_entityref(self, tag):
        pass

    def handle_charref(self, tag):
        pass

    def handle_comment(self, tag):
        pass

    def handle_decl(self, tag):
        pass

    def handle_pi(self, tag):
        pass

    def handle_unknown_decl(self, tag):
        pass


class UploadResultsError(Exception):
    pass


class UploadResults:
    """Submit results submission files to the ECF."""

    def __init__(self):
        """Build the user interface."""
        root = tkinter.Tk()
        root.wm_title("Upload ECF Results Submission File")
        # root.wm_resizable(width=tkinter.FALSE, height=tkinter.FALSE)
        frame = tkinter.ttk.Frame(master=root)
        frame.pack(expand=tkinter.TRUE, fill=tkinter.BOTH)
        frame.grid_rowconfigure(0)
        frame.grid_rowconfigure(1)
        frame.grid_rowconfigure(2)
        frame.grid_rowconfigure(3)
        frame.grid_rowconfigure(4)
        frame.grid_rowconfigure(5)
        frame.grid_rowconfigure(6)
        frame.grid_rowconfigure(7)
        frame.grid_rowconfigure(8)
        frame.grid_rowconfigure(9, weight=1)
        frame.grid_columnconfigure(0)
        frame.grid_columnconfigure(1, weight=1)
        tkinter.ttk.Label(master=frame, text="ECF URL").grid(
            column=0, row=0, padx=3
        )
        tkinter.ttk.Label(master=frame, text="File Name").grid(
            column=0, row=1, padx=3
        )
        tkinter.ttk.Label(master=frame, text="User Name").grid(
            column=0, row=2, padx=3
        )
        tkinter.ttk.Label(master=frame, text="Password").grid(
            column=0, row=3, padx=3
        )
        urlname = tkinter.StringVar()
        url_entry = tkinter.ttk.Entry(master=frame)
        url_entry["textvariable"] = urlname
        filename = tkinter.StringVar()
        fn_entry = tkinter.ttk.Entry(master=frame)
        fn_entry["textvariable"] = filename
        username = tkinter.StringVar()
        un_entry = tkinter.ttk.Entry(master=frame)
        un_entry["textvariable"] = username
        password = tkinter.StringVar()
        pw_entry = tkinter.ttk.Entry(master=frame, show="*")
        pw_entry["textvariable"] = password
        email_graders = tkinter.IntVar(value=1)
        eg_option = tkinter.ttk.Checkbutton(
            master=frame, variable=email_graders
        )
        tolerate_surname_mismatch = tkinter.IntVar()
        tsm_option = tkinter.ttk.Checkbutton(
            master=frame, variable=tolerate_surname_mismatch
        )
        tolerate_identical_games = tkinter.IntVar()
        tig_option = tkinter.ttk.Checkbutton(
            master=frame, variable=tolerate_identical_games
        )
        report_only = tkinter.IntVar(value=1)
        ro_option = tkinter.ttk.Checkbutton(master=frame, variable=report_only)
        text = tkinter.Text(master=frame, wrap=tkinter.WORD)
        scrollbar = tkinter.ttk.Scrollbar(
            master=text, orient=tkinter.VERTICAL, command=text.yview
        )
        text.configure(yscrollcommand=scrollbar.set)
        text.tag_configure(_AVOID_SCROLLBAR, rmargin=20)
        scrollbar.pack(side=tkinter.RIGHT, fill=tkinter.Y)
        url_entry.grid(column=1, row=0, sticky=tkinter.NSEW, padx=3, pady=3)
        fn_entry.grid(column=1, row=1, sticky=tkinter.NSEW, padx=3, pady=3)
        un_entry.grid(column=1, row=2, sticky=tkinter.NSEW, padx=3, pady=3)
        pw_entry.grid(column=1, row=3, sticky=tkinter.NSEW, padx=3, pady=3)
        eg_option.grid(column=0, row=4)
        tkinter.ttk.Label(
            master=frame,
            text="Email output to results officer named in submission file.",
        ).grid(column=1, row=4, sticky=tkinter.W, padx=3)
        tsm_option.grid(column=0, row=5)
        tkinter.ttk.Label(
            master=frame,
            text=" ".join(
                (
                    "Submission is allowed if submitted player name",
                    "does not match ECF database player name.",
                )
            ),
        ).grid(column=1, row=5, sticky=tkinter.W, padx=3)
        tig_option.grid(column=0, row=6)
        tkinter.ttk.Label(
            master=frame,
            text="Submission is allowed if identical games are present.",
        ).grid(column=1, row=6, sticky=tkinter.W, padx=3)
        ro_option.grid(column=0, row=7, padx=3)
        tkinter.ttk.Label(
            master=frame,
            text=" ".join(
                (
                    "Check and report submission,",
                    "but do not commit to ECF database.",
                )
            ),
        ).grid(column=1, row=7, sticky=tkinter.W, padx=3)
        tkinter.ttk.Label(master=frame, text=_START_TEXT).grid(
            column=0, row=8, columnspan=2, pady=3
        )
        text.grid(
            column=0, row=9, sticky=tkinter.NSEW, columnspan=2, padx=3, pady=3
        )
        menu = tkinter.Menu(master=frame, tearoff=False)
        menu.add_separator()
        menu.add_command(
            label="Set Default Live Upload URL",
            command=self.set_default_live_upload_url,
            accelerator="Alt F3",
        )
        menu.add_command(
            label="Set Default Test Upload URL",
            command=self.set_default_test_upload_url,
            accelerator="Alt F5",
        )
        menu.add_command(
            label="Select Submission File",
            command=self.select_submission_file,
            accelerator="Alt F7",
        )
        menu.add_separator()
        menu.add_command(
            label="Upload Submission File",
            command=self.upload_results_submission,
            accelerator="Alt F9",
        )
        menu.add_separator()
        menu.add_command(
            label="Save Response",
            command=self.save_response,
            accelerator="Alt F11",
        )
        menu.add_separator()
        menu.add_command(
            label="Save Session Log",
            command=self.save_session_log,
            accelerator="Alt F12",
        )
        menu.add_separator()
        self.menu = menu
        self._bind_for_scrolling_only(text)
        for w in (root, frame, text):
            for wc in w.winfo_children():
                wc.bind("<ButtonPress-3>", self.show_menu)
        un_entry.focus_set()
        self.root = root
        self.url_entry = url_entry
        self.fn_entry = fn_entry
        self.un_entry = un_entry
        self.pw_entry = pw_entry
        self.eg_option = eg_option
        self.tsm_option = tsm_option
        self.tig_option = tig_option
        self.ro_option = ro_option
        self.text = text
        self.urlname = urlname
        self.filename = filename
        self.username = username
        self.password = password
        self.email_graders = email_graders
        self.tolerate_surname_mismatch = tolerate_surname_mismatch
        self.tolerate_identical_games = tolerate_identical_games
        self.report_only = report_only
        self.most_recent_response = None

    # Put the scrollbar next to, not in, the Text widget and lose the tags.
    def insert_text(self, text):
        """Insert text at end of action log widget."""
        self.text.insert(tkinter.END, text, *(_AVOID_SCROLLBAR,))

    def show_menu(self, event=None):
        """Show the popup menu for widget."""
        self.menu.tk_popup(*event.widget.winfo_pointerxy())

    def select_submission_file(self, event=None):
        """Select a results submission file."""
        localfilename = tkinter.filedialog.askopenfilename(
            parent=self.text, title="Browse File", initialdir="~"
        )
        if not localfilename:
            return
        self.filename.set(localfilename)
        try:
            rft = open(localfilename).read()
        except Exception as exc:
            tkinter.messagebox.showinfo(
                title="Open Submission File", message=str(exc)
            )
            return
        if not self.is_results_file_text_sane(rft):
            message = "".join(
                (localfilename, " cannot be a valid submission file")
            )
            self.insert_text(message + ".\n\n")
            tkinter.messagebox.showinfo(
                title="Submission File", message=message
            )
        else:
            self.insert_text(
                "".join(
                    (
                        localfilename,
                        " may be a valid submission file: a few mandatory items",
                        " are present.\n\n",
                    )
                )
            )

    def set_default_live_upload_url(self, event=None):
        """Open or download a zip file by URL."""
        self.urlname.set(_DEFAULT_LIVE_URL)

    def set_default_test_upload_url(self, event=None):
        """Open or download a zip file by URL."""
        self.urlname.set(_DEFAULT_TEST_URL)

    def upload_results_submission(self, event=None):
        """Upload a file of results with curl or Requests package.

        curl, if available, is run by subprocess.run and the code snippet
        from Steve Bush is used if Requests package has to be used.

        """
        try:
            urlp = urllib.parse.urlparse(self.urlname.get())
        except ValueError as exc:
            tkinter.messagebox.showinfo(title="Invalid URL", message=str(exc))
            return
        if not urlp.scheme and not urlp.netloc and not urlp.path:
            tkinter.messagebox.showinfo(
                title="Invalid URL",
                message="".join(
                    (
                        "URL not present ",
                        "(eg: https://www.ecfrating.org.uk/v2/submit/)",
                    )
                ),
            )
            return
        if not urlp.scheme:
            tkinter.messagebox.showinfo(
                title="Incomplete URL",
                message="URL scheme missing (eg: https)",
            )
            return
        if not urlp.netloc:
            tkinter.messagebox.showinfo(
                title="Incomplete URL",
                message="URL address missing (eg: www.ecfrating.org.uk)",
            )
            return
        if not urlp.path:
            tkinter.messagebox.showinfo(
                title="Incomplete URL",
                message="URL path missing (eg: /v2/submit/)",
            )
            return
        localfilename = self.filename.get()
        try:
            rft = open(localfilename).read()
        except Exception as exc:
            self.insert_text(str(exc) + ".\n\n")
            tkinter.messagebox.showinfo(
                title="Open Submission File", message=str(exc)
            )
            return
        if not self.is_results_file_text_sane(rft):
            message = "".join(
                (localfilename, " cannot be a valid submission file")
            )
            self.insert_text(message + ".\n\n")
            tkinter.messagebox.showinfo(
                title="Submission File", message=message
            )
            return
        if not self.username.get():
            message = "No user name supplied"
            self.insert_text(message + ".\n\n")
            tkinter.messagebox.showinfo(title="User Name", message=message)
            return
        if not self.password.get():
            message = "No password supplied"
            self.insert_text(message + ".\n\n")
            tkinter.messagebox.showinfo(title="Password", message=message)
            return
        if not curl and not requests:
            message = "No method of submitting results is available"
            self.insert_text(
                "".join(
                    (
                        message,
                        ".\n\nThe 'curl' utility is not available on your ",
                        "system and the 'Requests' package (from Python ",
                        "Packages) is not installed.  One of these is ",
                        "required.\n\n",
                    )
                )
            )
            tkinter.messagebox.showinfo(title="Upload Method", message=message)
            return
        if self.urlname.get() == _DEFAULT_LIVE_URL:
            if not tkinter.messagebox.askyesno(
                title="Upload Results File",
                message="Do you want to upload to the live ECF database?",
            ):
                return
        if curl:
            args = ["curl", "-k", "-L"]
            args.append("-F username=" + self.username.get())
            args.append("-F password=" + self.password.get())
            args.append("-F uploaded_file=@" + self.filename.get())
            if self.email_graders.get():
                args.append("-F email_graders=")
            if self.tolerate_surname_mismatch.get():
                args.append("-F tolerate_surname_mismatch=")
            if self.tolerate_identical_games.get():
                args.append("-F tolerate_identical_games=")
            if self.report_only.get():
                args.append("-F report_only=")
            args.append(self.urlname.get())
            cp = subprocess.run(args, capture_output=True)
            if cp.returncode:
                message = "".join(
                    (
                        "'curl' exited with non-zero returncode '",
                        str(cp.returncode),
                        "'",
                    )
                )
                self.insert_text(
                    "".join(
                        (
                            message,
                            ".\nAssume upload of ",
                            self.filename.get(),
                            " failed.\n\nError output:\n",
                        )
                    )
                )
                self.insert_text(cp.stderr)
                tkinter.messagebox.showinfo(
                    title="Upload Error", message=message
                )
                self.password.set("")
                return
            message = "".join(
                ("'curl' exited with returncode '", str(cp.returncode), "'")
            )
            self.insert_text(
                "".join(
                    (
                        message,
                        ".\n\nResponse to upload of ",
                        self.filename.get(),
                        " is:\n",
                    )
                )
            )
            try:
                text = cp.stdout.decode("utf-8")
            except UnicodeDecodeError:
                try:
                    text = cp.stdout.decode("ascii")
                except UnicodeDecodeError:
                    text = cp.stdout.decode("iso-8859-1")
            self.most_recent_response = self.process_response(text)
            tkinter.messagebox.showinfo(
                title="Upload Response", message=message
            )
            self.password.set("")
            return
        if requests:
            postdata = {
                "username": self.username.get(),
                "password": self.password.get(),
            }
            if self.email_graders.get():
                postdata["email_graders"] = "on"
            if self.tolerate_surname_mismatch.get():
                postdata["tolerate_surname_mismatch"] = "on"
            if self.tolerate_identical_games.get():
                postdata["tolerate_identical_games"] = "on"
            if self.report_only.get():
                postdata["report_only"] = "on"
            filename = self.filename.get()
            try:
                of = open(filename, "rb")
            except Exception as exc:
                self.insert_text(str(exc) + ".\n\n")
                tkinter.messagebox.showinfo(
                    title="Submit Results File", message=str(exc)
                )
                self.password.set("")
                return
            filedata = {"uploaded_file": (filename, of)}
            try:
                response = requests.post(
                    self.urlname.get(), data=postdata, files=filedata
                )
            except Exception as exc:
                self.insert_text(str(exc) + ".\n\n")
                tkinter.messagebox.showinfo(
                    title="Submit Results File", message=str(exc)
                )
                self.password.set("")
                return
            message = "".join(
                (
                    "Response to upload request is status_code '",
                    str(response.status_code),
                    "'",
                )
            )
            self.insert_text(
                "".join(
                    (
                        message,
                        ".\n\nResponse to upload of ",
                        self.filename.get(),
                        " is:\n",
                    )
                )
            )
            self.most_recent_response = self.process_response(response.text)
            tkinter.messagebox.showinfo(
                title="Upload Response", message=message
            )
            self.password.set("")
            return
        self.password.set("")

    def save_response(self, event=None):
        """Save response to a successful upload."""
        if self.most_recent_response is None:
            tkinter.messagebox.showinfo(
                title="Save Response", message="No response available to save"
            )
            return
        if self.most_recent_response.responsestring is None:
            tkinter.messagebox.showinfo(
                title="Save Response",
                message="No data available in response to save",
            )
            return
        if (
            self.most_recent_response.submission_file_name
            != self.filename.get()
        ):
            if not tkinter.messagebox.askyesno(
                title="Save Response",
                message="".join(
                    (
                        "The current response is for\n\n",
                        self.most_recent_response.submission_file_name,
                        "\n\nnot the one shown in 'File Name'.\n\n",
                        "Do you want to save the current response?",
                    )
                ),
            ):
                return
        title = " ".join(
            (
                "Save Response for",
                os.path.basename(
                    self.most_recent_response.submission_file_name
                ),
            )
        )
        filename = tkinter.filedialog.asksaveasfilename(title=title)
        if not filename:
            return
        of = open(filename, mode="w")
        try:
            of.write(self.most_recent_response.responsestring)
        except Exception as exc:
            message = "Error while saving response"
            self.insert_text(
                "".join(
                    (
                        message,
                        " for ",
                        self.most_recent_response.submission_file_name,
                        ".\n\nReported exception is:\n",
                        str(exc),
                    )
                )
            )
            tkinter.messagebox.showinfo(title=title, message=message)
            return
        finally:
            of.close()
        self.insert_text(
            "".join(
                (
                    "Response for ",
                    self.most_recent_response.submission_file_name,
                    " saved in ",
                    filename,
                    ".\n\n",
                )
            )
        )

    def save_session_log(self, event=None):
        """Save tthe visible session log."""
        title = "Save Session Log"
        filename = tkinter.filedialog.asksaveasfilename(title=title)
        if not filename:
            return
        self.insert_text("".join(("Session Log saved in ", filename, ".\n\n")))
        of = open(filename, mode="w")
        try:
            of.write(self.text.get("1.0", tkinter.END))
        except Exception as exc:
            message = "Error while saving Session Log"
            self.insert_text(
                "".join(
                    (
                        message,
                        " in ",
                        filename,
                        "upload_results.py\n\nReported exception is:\n",
                        str(exc),
                    )
                )
            )
            tkinter.messagebox.showinfo(title=title, message=message)
            return
        finally:
            of.close()

    def is_results_file_text_sane(self, text):
        """Return True if text could be a valid ECF results submission.

        Verify a few mandatory fields are present.

        """
        te = [e.strip() for e in text.split("#")]
        event_details = False
        event_code = False
        submission_index = False
        event_name = False
        results_officer_address = False
        for t in te:
            t = t.split("=", 1)[0]
            if t == _EVENT_DETAILS:
                event_details = True
            elif t == _EVENT_CODE:
                event_code = True
            elif t == _SUBMISSION_INDEX:
                submission_index = True
            elif t == _EVENT_NAME:
                event_name = True
            elif t == _RESULTS_OFFICER_ADDRESS:
                results_officer_address = True
            else:
                continue
            if (
                event_details
                and event_code
                and submission_index
                and event_name
                and results_officer_address
            ):
                return True
        return False

    def process_response(self, response):
        fb = Feedback()
        fb.submission_file_name = self.filename.get()
        fb.responsestring = response
        fb.feed(response)
        fb.insert_whitespace()
        fb.find_player_lists()
        self.insert_text("\n\n")
        if (
            fb.feedbacknumbers is None
            or fb.feedbackplayers is None
            or fb.submissionpins is None
            or fb.submissionplayers is None
        ):
            message = "".join(
                (
                    "The upload is assumed to have failed because the player ",
                    "lists cannot be found",
                )
            )
            self.insert_text(message + ".\n\n")
            tkinter.messagebox.showinfo(
                title="Upload Response", message=message
            )
            return fb
        self.insert_text(
            "List of players generated in response to submission\n\n"
        )
        self.insert_text(fb.feedbackplayers[0])
        self.insert_text("\n")
        for n, p in zip(fb.feedbacknumbers, fb.feedbackplayers[1:]):
            self.insert_text("".join((n, p, "\n")))
        self.insert_text("\n\nList of players in the submission\n\n")
        self.insert_text(fb.submissionplayers[0])
        for n, p in zip(fb.submissionpins, fb.submissionplayers[1:]):
            self.insert_text("".join((n, p)))
        self.insert_text(
            "".join(
                (
                    "\n\nThe submission PINs and feedback players in display order ",
                    "match up as follows:\n\n",
                )
            )
        )
        for n, s, p in zip(
            fb.feedbacknumbers, fb.submissionpins, fb.feedbackplayers[1:]
        ):
            self.insert_text(
                "".join((n.strip(), "\t", s[1:], "\t\t", p, "\n"))
            )
        self.insert_text("\n\n")
        return fb

    def _bind_for_scrolling_only(self, widget):
        widget.bind("<KeyPress>", "break")
        widget.bind("<Home>", "return")
        widget.bind("<Left>", "return")
        widget.bind("<Up>", "return")
        widget.bind("<Right>", "return")
        widget.bind("<Down>", "return")
        widget.bind("<Prior>", "return")
        widget.bind("<Next>", "return")
        widget.bind("<End>", "return")


if __name__ == "__main__":

    UploadResults().root.mainloop()
